from hwt.simulator.agentBase import AgentBase
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy


class StructIntfAgent(AgentBase):
    """
    Agent for StructIntf inteface

    :summary: only purpose is to instantiate agents for child interfaces
    """
    def __init__(self, intf):
        AgentBase.__init__(self, intf)
        # if interface is InterfaceProxy and this proxy is container of another array
        # we generate agents on items in array
        # otherwise generate agents for each subinterface

        for intf in intf._interfaces:
            if intf._isInterfaceArray():
                agCls = intf[0]._getSimAgent()
                for p in intf:
                    p._ag = agCls(p)
            else:
                intf._ag = intf._getSimAgent()(intf)

    def getMonitors(self):
        for intf in self.intf._interfaces:
            if intf._arrayElemCache or (isinstance(intf, InterfaceProxy) and intf._origIntf._arrayElemCache):
                for p in intf:
                    yield from p._ag.getMonitors()
            else:
                yield from intf._ag.getMonitors()

    def getDrivers(self):
        for intf in self.intf._interfaces:
            if intf._arrayElemCache or (isinstance(intf, InterfaceProxy) and intf._origIntf._arrayElemCache):
                for p in intf:
                    yield from p._ag.getDrivers()
            else:
                yield from intf._ag.getDrivers()



from hwt.hdlObjects.constants import DIRECTION
from hwt.hdlObjects.typeShortcuts import vecT
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BIT
from hwt.interfaces.signalOps import SignalOps
from hwt.synthesizer.interfaceLevel.interface import Interface
from hwt.synthesizer.param import Param, evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


D = DIRECTION


class Signal(SignalOps, Interface):
    """
    Basic wire interface
    """
    def __init__(self,
                 masterDir=D.OUT,
                 asArraySize=None,
                 dtype=BIT,
                 loadConfig=True):
        super().__init__(masterDir=masterDir,
                         asArraySize=asArraySize,
                         loadConfig=loadConfig)
        self._dtype = dtype

    def _injectMultiplerToDtype(self):
        """
        Make signal wider, used when there is an array of signals stored in one wider signal
        """
        t = self._dtype
        factor = self._widthMultiplier
        if t == BIT:
            newT = vecT(factor)
        elif isinstance(t, Bits):
            w = t.width
            if isinstance(w, int):
                newW = factor * w
            elif isinstance(w, RtlSignalBase):
                # both Param or factor Value
                newW = w * factor
            elif isinstance(factor, RtlSignalBase):
                # w is Value
                newW = factor * w
            else:
                # both Value
                newW = w.clone()
                newW.val *= factor.val
            newT = vecT(newW)
        else:
            raise TypeError("Can not multiply width of type %r" % (repr(t),))

        self._dtype = newT

    def _getSimAgent(self):
        from hwt.interfaces.agents.signal import SignalAgent
        return SignalAgent


def VectSignal(width,
               signed=None,
               masterDir=D.OUT,
               asArraySize=None,
               loadConfig=True):
    """
    Create basic :class:`.Signal` interface where type is vector
    """
    return Signal(masterDir,
                  asArraySize,
                  vecT(width, signed),
                  loadConfig)


class Clk(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as clock signal
    """
    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Clk
        return IP_Clk

    def _getSimAgent(self):
        from hwt.interfaces.agents.clk import OscilatorAgent
        return OscilatorAgent


class Rst(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal
    """
    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst
        return IP_Rst

    def _getSimAgent(self):
        from hwt.interfaces.agents.rst import PullDownAgent
        return PullDownAgent


class Rst_n(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal with negative polarity (active in 0)
    """
    def _signalsForInterface(self,
                             context,
                             prefix='',
                             typeTransform=None):
        sigs = Signal._signalsForInterface(self,
                                           context,
                                           prefix,
                                           typeTransform=typeTransform)
        for s in sigs:
            s.negated = True
        return sigs

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_Rst_n
        return IP_Rst_n

    def _getSimAgent(self):
        from hwt.interfaces.agents.rst import PullUpAgent
        return PullUpAgent


class VldSynced(Interface):
    """
    Interface data+valid signal, if vld=1 then data are valid and slave should accept them
    """
    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = s(dtype=vecT(self.DATA_WIDTH))
        self.vld = s()

    def _getSimAgent(self):
        from hwt.interfaces.agents.vldSynced import VldSyncedAgent
        return VldSyncedAgent


class RdSynced(Interface):
    """
    Interface data+ready signal, if rd=1 then slave has read data and master should actualize data
    """
    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = s(dtype=vecT(self.DATA_WIDTH))
        self.rd = s(masterDir=D.IN)

    def _getSimAgent(self):
        from hwt.interfaces.agents.rdSynced import RdSyncedAgent
        return RdSyncedAgent


class Handshaked(VldSynced):
    """
    Interface data+ready+valid signal, if rd=1 slave is ready to accept data, if vld=1 master is sending data,
    if rd=1 and vld=1 then data is transfered otherwise master and slave has to wait on each other
    :attention: one rd/vld is set it must not go down until transaction is made
    """
    def _declr(self):
        super()._declr()
        self.rd = s(masterDir=D.IN)

    def _getSimAgent(self):
        from hwt.interfaces.agents.handshaked import HandshakedAgent
        return HandshakedAgent


class HandshakeSync(Interface):
    """
    Only synchronization interface, like vld+rd signal with meaning like in :class:`.Handshaked` interface
    
    :ivar rd: when high slave is ready to receive data
    :ivar vld: when high master is sending data to slave

    transaction happens when both ready and valid are high

    """
    def _declr(self):
        self.vld = s()
        self.rd = s(masterDir=D.IN)

    def _getSimAgent(self):
        from hwt.interfaces.agents.handshaked import HandshakeSyncAgent
        return HandshakeSyncAgent


class ReqDoneSync(Interface):
    """
    Synchronization interface, if req=1 slave begins operation and when its done it asserts done=1 for one clk tick
    req does not need to stay high
    """
    def _declr(self):
        self.req = s()
        self.done = s(masterDir=D.IN)


class BramPort_withoutClk(Interface):
    """
    Basic BRAM port
    """
    def _config(self):
        self.ADDR_WIDTH = Param(32)
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.addr = s(dtype=vecT(self.ADDR_WIDTH))
        self.din = s(dtype=vecT(self.DATA_WIDTH))
        self.dout = s(masterDir=D.IN, dtype=vecT(self.DATA_WIDTH))
        self.en = s()
        self.we = s()

    def _getIpCoreIntfClass(self):
        from hwt.serializer.ip_packager.interfaces.std import IP_BlockRamPort
        return IP_BlockRamPort

    def _getSimAgent(self):
        from hwt.interfaces.agents.bramPort import BramPort_withoutClkAgent
        return BramPort_withoutClkAgent

    def _getWordAddrStep(self):
        """
        :return: size of one word in unit of address
        """
        return 1

    def _getAddrStep(self):
        """
        :return: how many bits is one unit of address (f.e. 8 bits for  char * pointer,
             36 for 36 bit bram)
        """
        return evalParam(self.DATA_WIDTH).val


class BramPort(BramPort_withoutClk):
    """
    BRAM port with it's own clk
    """
    def _declr(self):
        self.clk = s(masterDir=D.OUT)
        with self._associated(clk=self.clk):
            super()._declr()

        self._associatedClk = self.clk

    def _getSimAgent(self):
        from hwt.interfaces.agents.bramPort import BramPortAgent
        return BramPortAgent

    @classmethod
    def fromBramPort_withoutClk(cls, bramPort, clk):
        assert isinstance(bramPort, BramPort_withoutClk)
        assert isinstance(clk, Clk)
        self = cls()
        rp = self._replaceParam

        def setIntf(name, intf):
            setattr(self, name, intf)
            self._interfaces.append(intf)

        rp("ADDR_WIDTH", bramPort.ADDR_WIDTH)
        rp("DATA_WIDTH", bramPort.DATA_WIDTH)

        self._interfaces = []
        for iName in ["addr", "din", "dout", "en", "we"]:
            intf = getattr(bramPort, iName)
            setIntf(iName, intf)
        setIntf("clk", clk)

        return self


class FifoWriter(Interface):
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.en = s()
        self.wait = s(masterDir=D.IN)
        self.data = s(dtype=vecT(self.DATA_WIDTH))

    def _getSimAgent(self):
        from hwt.interfaces.agents.fifo import FifoWriterAgent
        return FifoWriterAgent


class FifoReader(FifoWriter):
    def _declr(self):
        super()._declr()
        self.en._masterDir = D.IN
        self.wait._masterDir = D.OUT

    def _getSimAgent(self):
        from hwt.interfaces.agents.fifo import FifoReaderAgent
        return FifoReaderAgent


class RegCntrl(Interface):
    """
    Register control interface, :class:`.Signal` for read, :class:`.VldSynced` for write
    """
    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.din = Signal(dtype=vecT(self.DATA_WIDTH), masterDir=D.IN)
        with self._paramsShared():
            self.dout = VldSynced()

    def _getSimAgent(self):
        from hwt.interfaces.agents.regCntrl import RegCntrlAgent
        return RegCntrlAgent

s = Signal



from hwt.simulator.agentBase import SyncAgentBase


class FifoReaderAgent(SyncAgentBase):

    def __init__(self, intf, allowNoReset=False):
        super(FifoReaderAgent, self).__init__(intf, allowNoReset)
        self.data = []
        self.readPending = False

    def monitor(self, s):
        intf = self.intf
        r = s.read
        if self.readPending:
            yield s.updateComplete
            d = s.read(intf.data)
            self.data.append(d)
            self.readPending = False

        if r(self.rst_n).val and self.enable:
            wait = r(intf.wait)
            assert wait.vldMask
            rd = not wait.val
            s.write(rd, intf.en)

            if rd:
                self.readPending = True
        else:
            s.write(0, intf.en)

    def driver(self, s):
        raise NotImplementedError()


class FifoWriterAgent(SyncAgentBase):

    def monitor(self, s):
        raise NotImplementedError()

    def driver(self, s):
        intf = self.intf
        w = s.write
        r = s.read

        if r(self.rst_n).val and self.data and self.enable:
            wait = r(intf.wait)
            assert wait.vldMask
            if not wait.val:
                d = self.data.pop(0)
                w(d, intf.data)
                w(1, intf.en)
                return

        w(None, intf.data)
        w(0, intf.en)



from hwt.hdlObjects.statements import ReturnCalled


class Function():
    """
    Container for hdl function
    """
    def __init__(self, name, returnT, ctx, params, _locals, exprList, isOperator=False):
        """class to store hdl function

        :param name: name of the function
        :param returnT: return type
        :param ctx: hdl context of this funtion
        :param params: list of argument signals
        :param _locals: local variables in this function
        :param exprList: list of expressions in body
        :param isOperator: is operator flag
        """
        self.name = name
        self.returnT = returnT
        self.ctx = ctx
        self.params = params
        self.locals = _locals
        self.exprList = exprList
        self.isOperator = isOperator

    def call(self, *args):
        for p, a in zip(self.params, args):
            a.updateTime = 0
            p._val = a

        try:
            for s in self.exprList:
                s.seqEval()
        except ReturnCalled as r:
            return r.val
        return None

    def staticEval(self):
        return self



from hwt.hdlObjects.types.integer import Integer


class SimInteger(Integer):
    @classmethod
    def getConvertor(cls):
        from hwt.simulator.types.simIntConversions import convertSimInteger__val
        return convertSimInteger__val


SIM_INT = SimInteger()
# create hdl integer value (for example integer value in vhdl)


def simHInt(val):
    return SIM_INT.fromPy(val)



from hwt.hdlObjects.types.hdlType import HdlType
from hwt.hdlObjects.value import Value
from hwt.serializer.serializerClases.indent import getIndent


class HStructField(object):
    def __init__(self, typ, name, info=None):
        assert isinstance(name, str) or name is None, name
        assert isinstance(typ, HdlType), typ
        self.name = name
        self.dtype = typ
        self.info = info

    def __hash__(self):
        return hash((self.dtype, self.name))


class HStruct(HdlType):
    """
    C-like structure type

    :ivar fields: tuple of HStructField instances in this struct
    :ivar name: name of this HStruct type
    :ivar isFixedSize: flag which tells if size of this HStruct can resolved or not,
        f.e. field with type of HStream can cause isFixedSize to become False
    :ivar valueCls: Class of value for this type as usual in HdlType implementations
    """
    def __init__(self, *template, name=None):
        """
        :param template: list of tuples (type, name) or HStructField objects
            name can be None it means that there is space in structure
        :param name: optional name used for debugging purposes
        """
        self.fields = []
        self.name = name
        fieldNames = []
        self.isFixedSize = True

        for f in template:
            try:
                field = HStructField(*f)
            except TypeError:
                field = f
            if not isinstance(field, HStructField):
                raise TypeError("Template for struct field %s is not in valid format" % repr(f))

            self.fields.append(field)
            if field.name is not None:
                fieldNames.append(field.name)

            t = field.dtype
            if self.isFixedSize:
                if (isinstance(t, HStruct) and not t.isFixedSize)\
                              or not hasattr(t, "bit_length"):
                    self.isFixedSize = False

        self.fields = tuple(self.fields)
        if self.isFixedSize:
            self.__bit_length_val = self.__bit_length()

        class StructVal(Value):
            __slots__ = fieldNames

            def __init__(self, val, typeObj):
                self._dtype = typeObj

                for f in self._dtype.fields:
                    if f.name is None:
                        continue
                    v = val.get(f.name, None)
                    if not isinstance(v, Value):
                        v = f.dtype.fromPy(v)
                    setattr(self, f.name, v)

            @classmethod
            def fromPy(cls, val, typeObj):
                self = cls(val, typeObj)
                return self

            def __repr__(self):
                buff = ["{"]
                for f in self._dtype.fields:
                    if f.name is not None:
                        val = getattr(self, f.name)
                        buff.append("    %s %r" % (f.name, val))
                buff.append("}")
                return "\n".join(buff)

        protectedNames = set(["clone", "staticEval", "fromPy", "_dtype"])
        usedNames = set(fieldNames)
        assert not protectedNames.intersection(usedNames), protectedNames.intersection(usedNames)

        self.valueCls = StructVal

    def bit_length(self):
        if self.isFixedSize:
            return self.__bit_length_val
        else:
            raise TypeError("Can not request bit_lenght on size which has not fixed size")

    def __bit_length(self):
        return sum(map(lambda f: f.dtype.bit_length(), self.fields))

    def getValueCls(self):
        return self.valueCls

    def sizeof(self):
        """
        get size of struct in bytes
        """
        s = 0
        for f in self.template:
            s += f.dtype.bit_length()

        if s % 8 == 0:
            return s // 8
        else:
            return s // 8 + 1

    def __hash__(self):
        return hash((self.name, self.fields))

    def __add__(self, other):
        """
        override of addition, merge struct into one
        """
        assert isinstance(other, HStruct)
        return HStruct(*self.fields, *other.fields)

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not none is used as a additional information about where
            on which address this type is stored (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and Array)
        """
        if self.name:
            name = self.name + " "
        else:
            name = ""
        
            
        myIndent = getIndent(indent)
        childIndent = getIndent(indent + 1)
        header = "%sstruct %s{" % (myIndent, name) 

        buff = [header, ]
        for f in self.fields:
            if withAddr is not None:
                addrTag = " // start:0x%x(bit) 0x%x(byte)" % (withAddr, withAddr // 8)
            else:
                addrTag = ""

            if f.name is None:
                buff.append("%s//%r empty space%s" % (childIndent, f.dtype, addrTag))
            else:
                buff.append("%s %s%s" % (f.dtype.__repr__(indent=indent + 1,
                                                        withAddr=withAddr,
                                                        expandStructs=expandStructs),
                                       f.name, addrTag))
            if withAddr is not None:
                withAddr += f.dtype.bit_length()

        buff.append("%s}" % (myIndent))
        return "\n".join(buff)



from collections import namedtuple

from hwt.hdlObjects.architecture import Architecture
from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.entity import Entity
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.statements import IfContainer, \
    SwitchContainer, WhileContainer, WaitStm
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BOOL, BIT
from hwt.hdlObjects.types.enum import Enum
from hwt.hdlObjects.value import Value
from hwt.hdlObjects.variables import SignalItem
from hwt.pyUtils.arrayQuery import arr_any, groupedby
from hwt.serializer.constants import SERI_MODE
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.nameScope import LangueKeyword, NameScope
from hwt.serializer.serializerClases.indent import getIndent
from hwt.serializer.serializerClases.mapExpr import MapExpr
from hwt.serializer.serializerClases.portMap import PortMap
from hwt.serializer.utils import maxStmId
from hwt.serializer.vhdl.ops import VhdlSerializer_ops
from hwt.serializer.vhdl.statements import VhdlSerializer_statements
from hwt.serializer.vhdl.types import VhdlSerializer_types
from hwt.serializer.vhdl.utils import VhdlVersion, vhdlTmplEnv
from hwt.serializer.vhdl.value import VhdlSerializer_Value
from hwt.synthesizer.interfaceLevel.unit import Unit
from hwt.synthesizer.param import getParam, evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


architectureTmpl = vhdlTmplEnv.get_template('architecture.vhd')
entityTmpl = vhdlTmplEnv.get_template('entity.vhd')
processTmpl = vhdlTmplEnv.get_template('process.vhd')
componentTmpl = vhdlTmplEnv.get_template('component.vhd')
componentInstanceTmpl = vhdlTmplEnv.get_template('component_instance.vhd')

VHLD_KEYWORDS = [
    "abs", "access", "across", "after", "alias", "all", "and", "architecture", "array",
    "assert", "attribute", "begin", "block", "body", "break", "bugger", "bus", "case",
    "component", "configuration", "constant", "disconnect", "downto", "end", "entity",
    "else", "elsif", "exit", "file", "for", "function", "generate", "generic", "group",
    "guarded", "if", "impure", "in", "inertial", "inout", "is", "label", "library", "limit",
    "linkage", "literal", "loop", "map", "mod", "nand", "nature", "new", "next", "noise",
    "nor", "not", "null", "of", "on", "open", "or", "others", "out", "package", "port",
    "postponed", "process", "procedure", "procedural", "pure", "quantity", "range",
    "reverse_range", "reject", "rem", "record", "reference", "register", "report", "return",
    "rol", "ror", "select", "severity", "shared", "signal", "sla", "sll", "spectrum", "sra",
    "srl", "subnature", "subtype", "terminal", "then", "through", "to", "tolerance", "transport",
    "type", "unaffected", "units", "until", "use", "variable", "wait", "with", "when", "while",
    "xnor", "xor"]


def freeze_dict(data):
    keys = sorted(data.keys())
    frozen_type = namedtuple(''.join(keys), keys)
    return frozen_type(**data)


def paramsToValTuple(unit):
    d = {}
    for p in unit._params:
        name = p.getName(unit)
        v = evalParam(p)
        d[name] = v
    return freeze_dict(d)


def onlyPrintDefaultValues(suggestedName, dtype):
    # [TODO] it is better to use RtlSignal
    s = SignalItem(suggestedName, dtype, virtualOnly=True)
    s.hidden = False
    serializedS = VhdlSerializer.SignalItem(s, onlyPrintDefaultValues, declaration=True)
    print(serializedS)
    return s


class VhdlSerializer(VhdlSerializer_Value, VhdlSerializer_ops, VhdlSerializer_types, VhdlSerializer_statements):
    VHDL_VER = VhdlVersion.v2002
    __keywords_dict = {kw: LangueKeyword() for kw in VHLD_KEYWORDS}

    @staticmethod
    def formater(s):
        return s

    fileExtension = '.vhd'

    @classmethod
    def getBaseNameScope(cls):
        s = NameScope(True)
        s.setLevel(1)
        s[0].update(cls.__keywords_dict)
        return s

    @classmethod
    def serializationDecision(cls, obj, serializedClasses, serializedConfiguredUnits):
        """
        Decide if this unit should be serialized or not eventually fix name to fit same already serialized unit

        :param serializedClasses: unitCls : unitobj
        :param serializedConfiguredUnits: (unitCls, paramsValues) : unitObj
            where paramsValues are named tuple name:value
        """
        isEnt = isinstance(obj, Entity)
        isArch = isinstance(obj, Architecture)
        if isEnt:
            unit = obj.origin
        elif isArch:
            unit = obj.entity.origin
        else:
            return True

        assert isinstance(unit, Unit)
        m = unit._serializerMode

        if m == SERI_MODE.ALWAYS:
            return True
        elif m == SERI_MODE.ONCE:
            if isEnt:
                try:
                    prevUnit = serializedClasses[unit.__class__]
                except KeyError:
                    serializedClasses[unit.__class__] = unit
                    obj.name = unit.__class__.__name__
                    return True

                obj.name = prevUnit._entity.name
                return False

            return serializedClasses[unit.__class__] is unit
        elif m == SERI_MODE.PARAMS_UNIQ:
            params = paramsToValTuple(unit)
            k = (unit.__class__, params)
            if isEnt:
                try:
                    prevUnit = serializedConfiguredUnits[k]
                except KeyError:
                    serializedConfiguredUnits[k] = unit
                    return True

                obj.name = prevUnit._entity.name
                return False

            return serializedConfiguredUnits[k] is unit
        elif m == SERI_MODE.EXCLUDE:
            if isEnt:
                obj.name = unit.__class__.__name__
            return False
        else:
            raise NotImplementedError("Not implemented serializer mode %r on unit %r" % (m, unit))

    @classmethod
    def asHdl(cls, obj, createTmpVarFn, indent=0):
        """
        Convert object to VHDL string

        :param obj: object to serialize
        :param createTmpVarFn: function (sugestedName, dtype) returns variable
            this function will be called to create tmp variables
        """
        if hasattr(obj, "asVhdl"):
            return obj.asVhdl(cls, createTmpVarFn, indent)
        elif isinstance(obj, RtlSignalBase):
            return cls.SignalItem(obj, createTmpVarFn, indent)
        elif isinstance(obj, Value):
            return cls.Value(obj, createTmpVarFn)
        else:
            try:
                serFn = getattr(cls, obj.__class__.__name__)
            except AttributeError:
                raise NotImplementedError("Not implemented for %s" % (repr(obj)))
            return serFn(obj, createTmpVarFn, indent)

    @classmethod
    def FunctionContainer(cls, fn):
        return fn.name

    @classmethod
    def Architecture(cls, arch, scope, indent=0):
        variables = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: x.name)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        arch.components.sort(key=lambda x: x.name)
        arch.componentInstances.sort(key=lambda x: x._name)

        def createTmpVarFn(suggestedName, dtype):
            raise NotImplementedError()

        for v in arch.variables:
            t = v._dtype
            # if type requires extra definition
            if isinstance(t, (Enum, Array)) and t not in extraTypes:
                extraTypes.add(v._dtype)
                extraTypes_serialized.append(cls.HdlType(t, createTmpVarFn, scope, declaration=True))

            v.name = scope.checkedName(v.name, v)
            serializedVar = cls.SignalItem(v, createTmpVarFn, declaration=True, indent=indent + 1)
            variables.append(serializedVar)

        for p in arch.processes:
            procs.append(cls.HWProcess(p, scope, indent=indent + 1))

        # architecture names can be same for different entities
        # arch.name = scope.checkedName(arch.name, arch, isGlobal=True)

        uniqComponents = list(map(lambda x: x[1][0], groupedby(arch.components, lambda c: c.name)))
        uniqComponents.sort(key=lambda c: c.name)
        components = list(map(lambda c: cls.Component(c, createTmpVarFn, indent + 1),
                              uniqComponents))

        componentInstances = list(map(lambda c: cls.ComponentInstance(c, createTmpVarFn, scope, indent + 1),
                                      arch.componentInstances))

        return architectureTmpl.render(
            indent=getIndent(indent),
            entityName=arch.getEntityName(),
            name=arch.name,
            variables=variables,
            extraTypes=extraTypes_serialized,
            processes=procs,
            components=components,
            componentInstances=componentInstances
            )

    @classmethod
    def comment(cls, comentStr):
        return "--" + comentStr.replace("\n", "\n--")

    @classmethod
    def Component(cls, entity, createTmpVar, indent=0):
        entity.ports.sort(key=lambda x: x.name)
        entity.generics.sort(key=lambda x: x.name)
        return componentTmpl.render(
                indent=getIndent(indent),
                ports=[cls.PortItem(pi, createTmpVar) for pi in entity.ports],
                generics=[cls.GenericItem(g, createTmpVar) for g in entity.generics],
                entity=entity
                )

    @classmethod
    def ComponentInstance(cls, entity, createTmpVarFn, scope, indent=0):
        portMaps = []
        for pi in entity.ports:
            pm = PortMap.fromPortItem(pi)
            portMaps.append(pm)

        genericMaps = []
        for g in entity.generics:
            gm = MapExpr(g, g._val)
            genericMaps.append(gm)

        if len(portMaps) == 0:
            raise Exception("Incomplete component instance")

        # [TODO] check component instance name
        return componentInstanceTmpl.render(
                indent=getIndent(indent),
                instanceName=entity._name,
                entity=entity,
                portMaps=[cls.PortConnection(x, createTmpVarFn) for x in portMaps],
                genericMaps=[cls.MapExpr(x, createTmpVarFn) for x in genericMaps]
                )

    @classmethod
    def Entity(cls, ent, scope, indent=0):
        ports = []
        generics = []
        ent.ports.sort(key=lambda x: x.name)
        ent.generics.sort(key=lambda x: x.name)

        def createTmpVarFn(suggestedName, dtype):
            raise NotImplementedError()

        ent.name = scope.checkedName(ent.name, ent, isGlobal=True)
        for p in ent.ports:
            p.name = scope.checkedName(p.name, p)
            ports.append(cls.PortItem(p, createTmpVarFn))

        for g in ent.generics:
            g.name = scope.checkedName(g.name, g)
            generics.append(cls.GenericItem(g, createTmpVarFn))

        entVhdl = entityTmpl.render(
                indent=getIndent(indent),
                name=ent.name,
                ports=ports,
                generics=generics
                )

        doc = ent.__doc__
        if doc and id(doc) != id(Entity.__doc__):
            doc = cls.comment(doc) + "\n"
            return doc + entVhdl
        else:
            return entVhdl

    @classmethod
    def condAsHdl(cls, cond, forceBool, createTmpVarFn):
        if isinstance(cond, RtlSignalBase):
            cond = [cond]
        else:
            cond = list(cond)
        if len(cond) == 1:
            c = cond[0]
            if not forceBool or c._dtype == BOOL:
                return cls.asHdl(c, createTmpVarFn)
            elif c._dtype == BIT:
                return "(" + cls.asHdl(c, createTmpVarFn) + ")=" + cls.BitLiteral(1, 1)
            elif isinstance(c._dtype, Bits):
                width = c._dtype.bit_length()
                return "(" + cls.asHdl(c, createTmpVarFn) + ")/=" + cls.BitString(0, width)
            else:
                raise NotImplementedError()
        else:
            return " AND ".join(map(lambda x: cls.condAsHdl(x, forceBool, createTmpVarFn), cond))

    @classmethod
    def GenericItem(cls, g, createTmpVarFn):
        s = "%s : %s" % (g.name, cls.HdlType(g._dtype, createTmpVarFn))
        if g.defaultVal is None:
            return s
        else:
            return "%s := %s" % (s, cls.Value(getParam(g.defaultVal).staticEval(), createTmpVarFn))

    @classmethod
    def PortConnection(cls, pc, createTmpVarFn):
        if pc.portItem._dtype != pc.sig._dtype:
            raise SerializerException("Port map %s is nod valid (types does not match)  (%s, %s)" % (
                      "%s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, createTmpVarFn)),
                      repr(pc.portItem._dtype), repr(pc.sig._dtype)))
        return " %s => %s" % (pc.portItem.name, cls.asHdl(pc.sig, createTmpVarFn))

    @classmethod
    def DIRECTION(cls, d):
        return d.name

    @classmethod
    def PortItem(cls, pi, createTmpVarFn):
        return "%s : %s %s" % (pi.name, cls.DIRECTION(pi.direction),
                               cls.HdlType(pi._dtype, createTmpVarFn))

    @classmethod
    def sensitivityListItem(cls, item, createTmpVarFn):
        if isinstance(item, Operator):
            item = item.ops[0]
        return cls.asHdl(item, createTmpVarFn)

    @classmethod
    def HWProcess(cls, proc, scope, indent=0):
        """
        Serialize HWProcess objects as VHDL

        :param scope: name scope to prevent name collisions
        """
        body = proc.statements
        extraVars = []
        extraVarsSerialized = []

        def createTmpVarFn(suggestedName, dtype):
            # [TODO] it is better to use RtlSignal
            s = SignalItem(None, dtype, virtualOnly=True)
            s.name = scope.checkedName(suggestedName, s)
            s.hidden = False
            serializedS = cls.SignalItem(s, createTmpVarFn, declaration=True)
            extraVars.append(s)
            extraVarsSerialized.append(serializedS)
            return s

        hasToBeVhdlProcess = extraVars or arr_any(body, lambda x: isinstance(x,
                                        (IfContainer, SwitchContainer, WhileContainer, WaitStm)))

        sensitivityList = sorted(map(lambda s: cls.sensitivityListItem(s, None), proc.sensitivityList))

        if hasToBeVhdlProcess:
            sIndent = indent + 1
        else:
            sIndent = indent
            
        statemets = [cls.asHdl(s, createTmpVarFn, indent=sIndent) for s in body]

        if hasToBeVhdlProcess:
            proc.name = scope.checkedName(proc.name, proc)

        extraVarsInit = []
        for s in extraVars:
            a = Assignment(s.defaultVal, s, virtualOnly=True)
            extraVarsInit.append(cls.Assignment(a, createTmpVarFn, indent=indent + 1))

        return processTmpl.render(
            indent=getIndent(indent),
            name=proc.name,
            hasToBeVhdlProcess=hasToBeVhdlProcess,
            extraVars=extraVarsSerialized,
            sensitivityList=", ".join(sensitivityList),
            statements=extraVarsInit + statemets
            )



from hwt.hdlObjects.types.struct import HStructField, HStruct


# [TODO] rename to BusFieldMeta
class BusFieldInfo(object):
    def __init__(self, access="rw", fieldInterface=None, disolveArray=False):
        """
        :param access: "r", "w" or "rw" describes access mode from bus side
        :param fieldInterface: interface for which this field was generated
        :param disolveArray: interpret this array interface as bunch of items
            instead of single memory space
        """
        assert access in ['r', 'w', 'rw']
        self.access = access
        self.fieldInterface = fieldInterface
        self.disolveArray = disolveArray


def HStruct_selectFields(structT, fieldsToUse):
    """
    Select fields from structure (rest will become spacing)
    
    :param structT: HStruct type instance
    :param fieldsToUse: dict {name:{...}} or set of names to select, dictionary is used to select nested fields
        in HStruct or HUnion fields (f.e. {"struct1": {"field1", "field2"}, "field3":{}} 
        will select field1 and 2 from struct1 and field3 from root)
    """

    template = []
    fieldsToUse = fieldsToUse
    foundNames = set()

    for f in structT.fields:
        name = None
        subfields = []

        if f.name is not None:
            try:
                if isinstance(fieldsToUse, dict):
                    subfields = fieldsToUse[f.name]
                    name = f.name
                else:
                    if f.name in fieldsToUse:
                        name = f.name
            except KeyError:
                name = None
        
        if name is not None and subfields:
            template.append(HStructField(HStruct_selectFields(f.dtype, subfields), name))
        else:
            template.append(HStructField(f.dtype, name))
        
        if f.name is not None:
            foundNames.add(f.name)

    if isinstance(fieldsToUse, dict):
        fieldsToUse = set(fieldsToUse.keys())
    assert fieldsToUse.issubset(foundNames)

    return HStruct(*template)




from hwt.synthesizer.exceptions import TypeConversionErr


class HdlType():

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash((self.name))

    def fromPy(self, v):
        return self.getValueCls().fromPy(v, self)

    def convert(self, sigOrVal, toType):
        if sigOrVal._dtype == toType:
            return sigOrVal

        try:
            c = self._convert
        except AttributeError:
            c = self.getConvertor()
            self._convertor = c

        return c(self, sigOrVal, toType)

    @classmethod
    def getConvertor(cls):
        return HdlType.defaultConvert

    def defaultConvert(self, sigOrVal, toType):
        raise TypeConversionErr("Conversion of %r of type \n%r to type %r is not implemented"
                                % (sigOrVal, self, toType))

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not none is used as a additional information about where
            on which address this type is stored (used only by HStruct)
        :param expandStructs: expand HStruct types (used by HStruct and Array)
        """
        return "<HdlType %s>" % (self.__class__.__name__)



from hwt.hdlObjects.typeShortcuts import vec
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.defs import BIT


def aplyIndexOnSignal(sig, dstType, index):
    if sig._dtype == BIT or dstType == BIT:
        return sig[index]
    elif isinstance(dstType, Bits):
        w = dstType.width
        return sig[(w * (index + 1)):(w * index)]
    else:
        raise NotImplementedError()


def fitTo(what, to):
    """
    Slice signal "what" to fit in "to"
    or
    extend "what" with zeros to same width as "to"

    little-endian impl.
    """

    whatWidth = what._dtype.bit_length()
    toWidth = to._dtype.bit_length()
    if toWidth == whatWidth:
        return what
    elif toWidth < whatWidth:
        # slice
        return what[toWidth:]
    else:
        # extend
        return vec(0, toWidth - whatWidth)._concat(what)



# select = map, groupBy = itertools.groupby
from types import GeneratorType
from itertools import zip_longest
from math import inf


class DuplicitValueExc(Exception):
    pass


class NoValueExc(Exception):
    pass


def distinctBy(iterable, fn):
    s = set()
    for i in iterable:
        r = fn(i)
        if r not in s:
            s.add(fn(i))
            yield i


def first(iterable, fn):
    for i in iterable:
        if fn(i):
            return i


def single(iterable, fn):
    found = False
    ret = None

    for i in iterable:
        if fn(i):
            if found:
                raise DuplicitValueExc(i)
            found = True
            ret = i
    if not found:
        raise NoValueExc()
    return ret


def arr_any(iterable, fn):
    for _ in where(iterable, fn):
        return True
    return False


def where(iterable, fn):
    for i in iterable:
        if fn(i):
            yield i


def last_iter(it):
    # Ensure it's an iterator and get the first field
    it = iter(it)
    prev = next(it)
    for item in it:
        # Lag by one item so I know I'm not at the end
        yield False, prev
        prev = item
    # Last item
    yield True, prev


def extendLen(arr, newLen, useValue=None):
    lenNow = len(arr)
    toAdd = newLen - lenNow
    assert toAdd > 0
    arr.extend([useValue for _ in range(toAdd)])


def indexUsigIs(iterable, item):
    i = 0
    for v in iterable:
        if v is item:
            return i
        i += 1


def groupedby(collection, fn):
    """
    This function does not needs initial sorting like itertools.groupby

    :attention: Order of pairs is not deterministic.
    """
    d = {}
    for item in collection:
        k = fn(item)
        try:
            arr = d[k]
        except KeyError:
            arr = []
            d[k] = arr
        arr.append(item)

    yield from d.items()


def split(arr, size):
    arr = list(arr)
    while len(arr) > size:
        pice = arr[:size]
        yield pice
        arr = arr[size:]
    yield arr


def flatten(iterables, level=inf):
    """
    Flatten nested lists, tuples, generators and maps
    
    :param level: maximum depth of flattening 
    """
    if level >= 0 and isinstance(iterables, (list, tuple, GeneratorType, map, zip)):
        level -= 1
        for i in iterables:
            yield from flatten(i, level=level)
    else:
        yield iterables


def grouper(n, iterable, padvalue=None):
    "grouper(3, 'abcdefg', 'x') --> ('a','b','c'), ('d','e','f'), ('g','x','x')"
    return zip_longest(*[iter(iterable)] * n, fillvalue=padvalue)



from hwt.hdlObjects.assignment import Assignment
from hwt.hdlObjects.operator import Operator
from hwt.hdlObjects.operatorDefs import isEventDependentOp
from hwt.hdlObjects.portItem import PortItem
from hwt.hdlObjects.statements import IfContainer, SwitchContainer
from hwt.hdlObjects.value import Value
from hwt.pyUtils.arrayQuery import where
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


def signalHasDriver(sig):
    for _ in walkSignalDrivers(sig):
        return True
    return False


def walkSignalDrivers(sig):
    def assign2Me(ep):
        if isinstance(ep, Assignment):
            return True
        elif isinstance(ep, PortItem) and ep.dst is sig:
            return True
        else:
            return None

    return where(sig.drivers, assign2Me)


def walkAllOriginSignals(sig, discovered=None):
    """
    Walk every signal which has no driver and is used as driver of this signal.
    Goal is walk every generic in static expr. evaluation.
    """
    if discovered is None:
        discovered = set()
    if isinstance(sig, Value):
        return
    if not isinstance(sig, RtlSignalBase):
        raise AssertionError("Expected only instances of signal, got: %s"
                             % (repr(sig)))
    if sig in discovered:
        return
    discovered.add(sig)

    if sig.drivers:
        for obj in sig.drivers:
            if isinstance(obj, Operator):
                for op in obj.ops:
                    if isinstance(op, RtlSignalBase):
                        yield from walkAllOriginSignals(op, discovered=discovered)
            elif isinstance(obj, Assignment):
                yield from walkAllOriginSignals(obj.src, discovered)
            else:
                raise TypeError("walkAllOriginSignals not implemented for %s" % (str(obj)))
    else:
        yield sig


class EventDependencyReached(Exception):
    """
    Exception which signalize that walkDriversInExpr
    has reached event dependent operator
    """
    def __init__(self, evOp):
        self.evOp = evOp


def walkDriversInExpr(expr, seenSet):
    """
    :return: generators of RtlSignal
        where signal is in expression and is not used in event dependent expression
    :raise EventDependencyReached: when this generator steps on event dependent operator
    :attention: if event operator is found in expression, only sensitivity EventDependencyReached is raised
        this may case some synthesis (Vivado, ISE, Quartus ...) to complain about sensitivity,
        but it will work
    """
    if isinstance(expr, (Value, Param)):
        pass
    elif isinstance(expr, RtlSignalBase):
        if expr in seenSet:
            pass
        else:
            seenSet.add(expr)

            if not expr.hidden:
                yield expr
                return

            try:
                op = expr.singleDriver()
            except MultipleDriversExc:
                yield expr
                return

            if not isinstance(op, Operator):
                yield expr
                return

            if isEventDependentOp(op.operator):
                raise EventDependencyReached(op)
            else:
                for operand in op.ops:
                    yield from walkDriversInExpr(operand, seenSet)

    else:
        raise TypeError(expr)


def discoverEventDependency(sig):
    """
    walk signals drivers and yields whose signals which are in some event operator
    """
    try:
        drivers = sig.drivers
    except AttributeError:
        return

    if len(drivers) == 1:
        d = drivers[0]
        if isinstance(d, Operator):
            if isEventDependentOp(d.operator):
                yield d.ops[0]
            else:
                for op in d.ops:
                    yield from discoverEventDependency(op)


def _discoverSensitivityForList(cond, seenSet):
    """
    Discover sensitivity for list of signals
    """
    # do not yield directly to let EventDependencyReached exception propagate first
    tmp = []
    # if true
    for c in cond:
        tmp.extend(walkDriversInExpr(c, seenSet))
    yield from tmp


def _discoverSensitivity(statement, seenSet, isTop):
    try:
        if isinstance(statement, Assignment):
            if statement.indexes:
                yield from _discoverSensitivityForList(statement.indexes, seenSet)
            yield from walkDriversInExpr(statement.src, seenSet)
        elif isinstance(statement, IfContainer):
            # if true
            yield from _discoverSensitivityForList(statement.cond, seenSet)
            for stm in statement.ifTrue:
                yield from _discoverSensitivity(stm, seenSet, True)

            # elifs
            for cond, stms in statement.elIfs:
                yield from _discoverSensitivityForList(cond, seenSet)
                for stm in stms:
                    yield from _discoverSensitivity(stm, seenSet, True)
            # else
            for stm in statement.ifFalse:
                yield from _discoverSensitivity(stm, seenSet, True)

        elif isinstance(statement, SwitchContainer):
            # do not yield directly to let EventDependencyReached exception propagate first
            tmp = list(walkDriversInExpr(statement.switchOn, seenSet))
            yield from tmp
            for cond, stms in statement.cases:
                # yield from walkDriversInExpr(cond, seenSet)
                for stm in stms:
                    yield from _discoverSensitivity(stm, seenSet, True)

            for stm in statement.default:
                yield from _discoverSensitivity(stm, seenSet, True)
        else:
            raise TypeError(statement)

    except EventDependencyReached as e:
        if isTop:
            yield e.evOp
        else:
            raise e


def discoverSensitivity(statement):
    """
    :return: generators of (sensitivity, signal)
        where sensitivity is member of SENSITIVITY enum
    """
    yield from _discoverSensitivity(statement, set(), True)
    


from random import Random

from hwt.serializer.serializerClases.indent import getIndent
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy


def valueHasChanged(valA, valB):
    return valA.val is not valB.val or valA.vldMask != valB.vldMask


def agent_randomize(agent, timeQuantum, seed):
    random = Random(seed)

    def randomEnProc(simulator):
        # small space at start to modify agents when they are inactive
        yield simulator.wait(timeQuantum / 4)
        while True:
            agent.enable = random.random() < 0.5
            delay = int(random.random() * timeQuantum)
            yield simulator.wait(delay)

    return randomEnProc


def pprintInterface(intf, prefix="", indent=0):
    """
    Pretty print interface
    """
    try:
        s = intf._sig
    except AttributeError:
        s = ""
    if s is not "":
        s = repr(s)
    
    print("".join([getIndent(indent), prefix, repr(intf._getFullName()), " ", s]))
    for i in intf._interfaces:
        if isinstance(intf, InterfaceProxy):
            assert isinstance(i, InterfaceProxy), (intf, i)
        pprintInterface(i, indent=indent + 1)
    
    if intf._arrayElemCache:
        assert len(intf) == len(intf._arrayElemCache)
        for i, p in enumerate(intf):
            pprintInterface(p, prefix="p%d:" % i, indent=indent + 1)
    
def _pprintAgents(intf, indent, prefix=""):
    if intf._ag is not None:
        print("%s%s%r" % (getIndent(indent), prefix, intf._ag))
    for i in intf._interfaces:
        _pprintAgents(i, indent + 1)

    if intf._arrayElemCache:
        assert len(intf) == len(intf._arrayElemCache)
        for i, p in enumerate(intf):
            _pprintAgents(p, indent + 1, prefix="p%d:" % i)
        
    
def pprintAgents(unitOrIntf, indent=0):
    """
    Pretty print agents
    """
    prefix = unitOrIntf._name + ":"
    for intf in unitOrIntf._interfaces:
        _pprintAgents(intf, indent, prefix)


from enum import Enum


class SERI_MODE(Enum):
    """
    Serializer mode for unit

    Modes are driving when unit should be serialized

    :cvar ALWAYS: always serialize as independent unit
    :cvar ONCE: serialize only once and all other occurrences of this unit replace with first serialized one
    :cvar PARAMS_UNIQ: serialize when params are unique, units with same params will be serialized as single unit
    :cvar EXCLUDE: exclude completely from serialization (useful for archetypes like LUT, GT, DPS block etc. which are already in vendor std. lib)
        cls.__name__ (name of class) is used for all component instances
    """
    ALWAYS = 0
    ONCE = 1
    PARAMS_UNIQ = 2
    EXCLUDE = 3



from math import ceil, floor, inf
import re

from hwt.hdlObjects.transactionPart import TransactionPart
from hwt.hdlObjects.types.array import Array
from hwt.hdlObjects.types.bits import Bits
from hwt.hdlObjects.types.struct import HStruct
from hwt.simulator.types.simBits import simBitsT
from hwt.bitmask import mask, selectBitRange, setBitRange


class FrameTemplate(object):
    """
    Frame template container for informations about frame,
    template which is used for resolving how data should be formated into words and frames
    on target interface

    :ivar _fieldToTPart: dictionary {HStructField: TransactionPart} to resolve this association,
        None by default, builded when packData is called and is not builded
    """
    __RE_RM_ARRAY_DOTS = re.compile("(\.\[)")

    def __init__(self, origin, wordWidth, startBitAddr, endBitAddr, transactionParts):
        """
        :param origin: instance of HType (usually HStruct) from which this FrameTemplate was generated from
        :param wordWidth: width of word on interface where this template should be used
        :param startBitAddr: bit offset where this frame starts
        :param endBitAddr: bit offset where this frame ends (bit index of first bit behind this frame)
        :param transactionParts: instances of TransactionPart which are parts of this frame
        """
        self.origin = origin
        self.wordWidth = wordWidth
        self.startBitAddr = startBitAddr
        self.endBitAddr = endBitAddr
        self.parts = transactionParts

        self._fieldToTPart = None
        
        for p in self.parts:
            p.parent = self
            assert p.startOfPart >= startBitAddr, (p, startBitAddr, self)
            assert p.endOfPart <= endBitAddr, (p, endBitAddr, self)

    @staticmethod
    def frameFromTransTmpl(transactionTmpl,
                           wordWidth,
                           maxPaddingWords=inf,
                           trimPaddingWordsOnStart=False,
                           trimPaddingWordsOnEnd=False):
        """
        Params same as framesFromTransTmpl
        """
        return next(FrameTemplate.framesFromTransTmpl(transactionTmpl,
                                                      wordWidth,
                                                      maxPaddingWords=maxPaddingWords,
                                                      trimPaddingWordsOnStart=trimPaddingWordsOnStart,
                                                      trimPaddingWordsOnEnd=trimPaddingWordsOnEnd))
    
    @staticmethod
    def framesFromTransTmpl(transactionTmpl,
                            wordWidth,
                            maxFrameLen=inf,
                            maxPaddingWords=inf,
                            trimPaddingWordsOnStart=False,
                            trimPaddingWordsOnEnd=False):
        """
        Convert transaction template into FrameTemplates

        :param transactionTmpl: transaction template used which are FrameTemplates created from
        :param wordWidth: width of data signal in target interface where frames will be used
        :param maxFrameLen: maximum length of frame, if exceeded another frame will be created
        :param maxPaddingWords: maximum of continual padding words in frame,
            if exceed frame is split and words are cut of
        :attention: if maxPaddingWords<inf trimPaddingWordsOnEnd or trimPaddingWordsOnStart has to be True
            to decide where padding should be trimmed
        :param trimPaddingWordsOnStart: trim padding from start of frame at word granularity
        :param trimPaddingWordsOnEnd: trim padding from end of frame at word granularity
        """
        isFirstInFrame = True
        partsPending = False

        startOfThisFrame = 0
        endOfThisFrame = maxFrameLen
        parts = []
        endOfPart = 0
        assert maxFrameLen > 0
        assert maxPaddingWords >= 0
        if maxPaddingWords < inf:
            assert trimPaddingWordsOnStart or trimPaddingWordsOnEnd

        def fullWordCnt(start, end):
            """Count of complete words between two addresses
            """
            startWIndex = start // wordWidth
            endWIndex = end // wordWidth

            assert startWIndex <= endWIndex, (start, end)

            return endWIndex - startWIndex

        for (base, end), tmpl in transactionTmpl.walkFlatten():
            startOfPart = base
            while startOfPart != end:
                assert startOfThisFrame % wordWidth == 0, startOfThisFrame
                # parts are always in single word
                if startOfPart == endOfThisFrame:
                    # cut off padding at end of frame
                    paddingWords = fullWordCnt(endOfPart, endOfThisFrame)
                    if trimPaddingWordsOnEnd and paddingWords > maxPaddingWords:
                        _endOfThisFrame = endOfThisFrame - paddingWords * wordWidth
                        # align end of frame to word
                        _endOfThisFrame = ceil(_endOfThisFrame / wordWidth) * wordWidth

                    yield FrameTemplate(transactionTmpl, wordWidth, startOfThisFrame, _endOfThisFrame, parts)

                    # prepare for start of new frame
                    parts = []
                    isFirstInFrame = True
                    partsPending = False
                    # start on new
                    startOfThisFrame = (endOfThisFrame // wordWidth) * wordWidth
                    endOfThisFrame = startOfThisFrame + maxFrameLen
                    endOfPart = endOfThisFrame

                    continue

                if isFirstInFrame:
                    partsPending = True
                    isFirstInFrame = False
                    # cut off padding at start of frame
                    paddingWords = fullWordCnt(startOfThisFrame, base)
                    if trimPaddingWordsOnStart and paddingWords > maxPaddingWords:
                        startOfThisFrame += paddingWords * wordWidth

                    endOfThisFrame = startOfThisFrame + maxFrameLen
                else:
                    padding = startOfPart - endOfPart
                    if trimPaddingWordsOnEnd and padding >= wordWidth:
                        # there is too much continual padding
                        endOfThisFrame = startOfPart
                        continue

                wordIndex = startOfPart // wordWidth
                endOfWord = wordWidth * (wordIndex + 1)

                endOfPart = min(endOfWord, end, endOfThisFrame)

                inFieldOffset = startOfPart - base
                p = TransactionPart(tmpl, startOfPart, endOfPart, inFieldOffset)
                parts.append(p)

                startOfPart = endOfPart

        if partsPending:
            endOfThisFrame = max(startOfPart, transactionTmpl.bitAddrEnd)
            # cut off padding at end of frame
            paddingWords = fullWordCnt(endOfPart, endOfThisFrame)
            if trimPaddingWordsOnEnd and paddingWords > maxPaddingWords:
                endOfThisFrame -= paddingWords * wordWidth
                # align end of frame to word
            endOfThisFrame = ceil(endOfThisFrame / wordWidth) * wordWidth

            yield FrameTemplate(transactionTmpl, wordWidth, startOfThisFrame, endOfThisFrame, parts)

    def _wordIndx(self, addr):
        """
        convert bit address to index of word where this address is
        """
        return floor(addr / self.wordWidth)

    def getWordCnt(self):
        """
        Get count of words in this frame
        """
        return ceil((self.endBitAddr - self.startBitAddr) / self.wordWidth)

    def walkWords(self, showPadding=False):
        """
        Walk enumerated words in this frame

        :attention: not all indexes has to be present, only words with items will be generated when not showPadding
        :param showPadding: padding TransactionParts are also present
        :return: generator of tuples (wordIndex, list of TransactionParts in this word)
        """
        wIndex = 0
        lastEnd = self.startBitAddr
        parts = []
        for p in self.parts:
            end = p.startOfPart
            if showPadding and end != lastEnd:
                # insert padding
                while end != lastEnd:
                    assert end >= lastEnd, (end, lastEnd)
                    endOfWord = (self._wordIndx(lastEnd) + 1) * self.wordWidth
                    endOfPadding = min(endOfWord, end)
                    _p = TransactionPart(None, lastEnd, endOfPadding, 0)
                    _p.parent = self
                    parts.append(_p)

                    if endOfPadding >= endOfWord:
                        yield (wIndex, parts)
                        wIndex += 1
                        parts = []

                    lastEnd = endOfPadding

            if self._wordIndx(lastEnd) != self._wordIndx(p.startOfPart):
                yield (wIndex, parts)

                wIndex += 1
                parts = []
                lastEnd = p.endOfPart

            parts.append(p)
            lastEnd = p.endOfPart
            if lastEnd % self.wordWidth == 0:
                yield (wIndex, parts)

                wIndex += 1
                parts = []

        if showPadding and (parts or lastEnd != self.endBitAddr or lastEnd % self.wordWidth != 0):
            end = (self._wordIndx(self.endBitAddr - 1) + 1) * self.wordWidth
            if showPadding:
                while end != lastEnd:
                    assert end >= lastEnd, (end, lastEnd)
                    endOfWord = lastEnd + self.wordWidth
                    endOfPadding = min(endOfWord, end)
                    _p = TransactionPart(None, lastEnd, endOfPadding, 0)
                    _p.parent = self
                    parts.append(_p)

                    if endOfPadding >= endOfWord:
                        yield (wIndex, parts)
                        wIndex += 1
                        parts = []

                    lastEnd = endOfPadding

            if parts:
                yield (wIndex, parts)

    @staticmethod
    def buildFieldToDataDict(dtype, data, res):
        # assert data is None or isinstance(data, dict)
        for f in dtype.fields:
            try:
                fVal = data[f.name]
            except KeyError:
                fVal = None
            
            if isinstance(f.dtype, Bits):
                if fVal is not None:
                    assert isinstance(fVal, int)
                    res[f] = fVal
            elif isinstance(f.dtype, HStruct):
                if fVal:
                    FrameTemplate.buildFieldToDataDict(f.dtype, fVal, res)
            elif isinstance(f.dtype, Array):
                if fVal:
                    # assert isinstance(fVal, class_or_tuple)
                    res[f] = fVal
        
        return res
    
    def packData(self, data):
        """
        Pack data into list of BitsVal of specified dataWidth
    
        :param data: dict of values for struct fields {fieldName: value}
    
        :return: list of BitsVal which are representing values of words
        """
        typeOfWord = simBitsT(self.wordWidth, None)
        fieldToVal = self._fieldToTPart
        if fieldToVal is None:
            fieldToVal = self._fieldToTPart = self.buildFieldToDataDict(self.origin.dtype, data, {})

        for _, transactionParts in self.walkWords(showPadding=True):
            actualVldMask = 0
            actualVal = 0
            for tPart in transactionParts:
                high, low = tPart.getBusWordBitRange()
                fhigh, flow = tPart.getFieldBitRange()
                if not tPart.isPadding:
                    val = fieldToVal.get(tPart.tmpl.origin, None)
                else:
                    val = None
    
                if val is None:
                    newBits = 0
                    vld = 0
                else:
                    newBits = selectBitRange(val, flow, fhigh - flow)
                    vld = mask(high - low) << low
    
                actualVal = setBitRange(actualVal, low, high - low, newBits)
                actualVldMask = setBitRange(actualVal, low, high - low, vld)
                    
            yield typeOfWord.getValueCls()(actualVal, typeOfWord, actualVldMask, -1)

    def __repr__getName(self, transactionPart, fieldWidth):
        if transactionPart.isPadding:
            return "X"*fieldWidth
        else:
            names = []
            tp = transactionPart.tmpl
            while tp is not None:
                try:
                    isArrayElm = isinstance(tp.parent.dtype, Array)
                except AttributeError:
                    isArrayElm = False

                if isArrayElm:
                    arr = transactionPart.tmpl.parent
                    arrS = arr.bitAddr
                    itemW = (arr.bitAddrEnd - arrS) // arr.itemCnt
                    s = transactionPart.startOfPart
                    indx = (s - arrS) // itemW
                    names.append("[%d]" % indx)
                else:
                    o = tp.origin
                    if o is None:
                        break
                    if o.name is not None:
                        names.append(o.name)
                tp = tp.parent

            # [HOTFIX] rm dots when indexing on array
            return self.__RE_RM_ARRAY_DOTS.sub("[", ".".join(reversed(names)))

    def __repr__word(self, index, width, padding, transactionParts):
        buff = ["{0: <{padding}}|".format(index, padding=padding)]
        DW = self.wordWidth

        for tp in reversed(transactionParts):
            percentOfWidth = tp.bit_length() / DW
            # -1 for ending |
            fieldWidth = max(0, int(percentOfWidth * width) - 1)
            assert fieldWidth >= 0

            # percentOffset = (tp.inFrameBitAddr % DW) / DW
            # offset = int(percentOffset * width)
            name = self.__repr__getName(tp, fieldWidth)
            buff.append('{0: ^{fieldWidth}}|'.format(name, fieldWidth=fieldWidth))

        return "".join(buff)

    def __repr__(self, scale=1):
        buff = []
        s = "<%s start:%d, end:%d" % (self.__class__.__name__, self.startBitAddr, self.endBitAddr)
        if not self.parts:
            return s + ">"
        buff.append(s)

        padding = 5
        DW = self.wordWidth
        width = int(DW * scale)

        buff.append('{0: <{padding}}{1: <{halfLineWidth}}{2: >{halfLineWidth}}'.format(
            "", DW - 1, 0, padding=padding, halfLineWidth=width // 2))
        line = '{0: <{padding}}{1:-<{lineWidth}}'.format(
            "", "", padding=padding, lineWidth=width + 1)
        buff.append(line)

        for w, transactionParts in self.walkWords(showPadding=True):
            buff.append(self.__repr__word(w, width, padding, transactionParts))

        buff.append(line)
        buff.append(">")
        return "\n".join(buff)



from hwt.bitmask import mask
from hwt.hdlObjects.types.hdlType import HdlType
from hwt.serializer.serializerClases.indent import getIndent


class Bits(HdlType):

    def __init__(self, width, forceVector=False, signed=None):
        """
        :param forceVector: use always hdl vector type
            (for example std_logic_vector(0 downto 0) instead of std_logic)
        """
        self.forceVector = forceVector

        if isinstance(width, int):
            assert width > 0
        else:
            w = width.staticEval()
            assert w._isFullVld() and w.val > 0

        self.signed = signed
        self.width = width

    def __eq__(self, other):
        return isinstance(other, Bits) and other.bit_length() == self.bit_length()\
            and self.signed == other.signed and self.forceVector == other.forceVector

    def __hash__(self):
        return hash((self.signed, self.bit_length(), self.forceVector))

    def all_mask(self):
        return mask(self.bit_length())

    def bit_length(self):
        if isinstance(self.width, int):
            return self.width
        else:
            return self.width.staticEval().val

    @classmethod
    def getConvertor(cls):
        from hwt.hdlObjects.types.bitsConversions import convertBits
        return convertBits

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdlObjects.types.bitsVal import BitsVal
            cls._valCls = BitsVal
            return cls._valCls

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not none is used as a additional information about where
            on which address this type is stored (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and Array)
        """
        from hwt.serializer.vhdl.serializer import VhdlSerializer, onlyPrintDefaultValues
        c = self.width
        if isinstance(c, int):
            constr = "%dbits" % c
        else:
            constr = VhdlSerializer.asHdl(self.width, onlyPrintDefaultValues) + \
                     (", %dbits" % self.bit_length())

        return "%s<HdlType %s, %s>" % (getIndent(indent),
                                       self.__class__.__name__,
                                       constr)



from hwt.simulator.agentBase import AgentBase


def isPulledUpTristateRising(intf, sim):
    """
    check if there is a rising edge on tristate interface with pull up (open drain)
    """
    t = sim.read(intf.t)
    assert t.vldMask
    o = sim.read(intf.o)
    assert o.vldMask
    assert not o.val, "This is open drain mode, ioblock would burn"

    return bool(t._onFallingEdge(sim.now))


class TristatePullUpAgent(AgentBase):
    def __init__(self, intf, onRisingCallback=None, onFallingCallback=None):
        AgentBase.__init__(self, intf)
        self.onRisingCallback = onRisingCallback
        self.onFallingCallback = onFallingCallback

    def monitor(self, sim):
        intf = self.intf
        t = sim.read(intf.t)
        o = sim.read(intf.o)

        if sim.now > 0:
            assert t.vldMask, sim.now
            assert o.vldMask
            assert not o.val, "This is open drain mode, ioblock would burn"

        if t.val:
            v = 0
        else:
            v = 1

        last = sim.read(intf.i).val
        sim.write(v, intf.i)
        if not last and v and self.onRisingCallback:
            self.onRisingCallback(sim)
        if not v and last and self.onFallingCallback:
            self.onFallingCallback(sim)

    def getMonitors(self):
        return [self.onTWriteCallback]

    def onTWriteCallback(self, sim):
        """
        Process for injecting of this callback loop into simulator
        """
        self.monitor(sim)
        self.intf.t._sigInside._writeCallbacks.append(self.onTWriteCallback)
        # no function just asset this function will be generator
        yield sim.wait(0)



from hwt.hdlObjects.constants import DIRECTION
from hwt.hdlObjects.types.struct import HStruct
from hwt.synthesizer.interfaceLevel.interface import Interface


class StructIntf(Interface):
    """
    Create dynamic interface based on HStruct description

    :ivar _fieldsToInterfaces: dictionary {field from HStruct template: sub interface for it}
    """
    def __init__(self, structT, instantiateFieldFn, masterDir=DIRECTION.OUT, asArraySize=None, loadConfig=True):
        """
        :param structT: HStruct instance used as template for this interface
        :param instantiateFieldFn: function(FieldTemplateItem instance) used to instantiate fields
            (is called only on fields which have different type than HStruct)
        """
        Interface.__init__(self, masterDir=masterDir, asArraySize=asArraySize, loadConfig=loadConfig)
        self._structT = structT
        self._instantiateFieldFn = instantiateFieldFn
        self._fieldsToInterfaces = {}

    def _declr(self):
        for field in self._structT.fields:
            # skip padding
            if field.name is not None:
                # generate interface based on struct field
                t = field.dtype
                if isinstance(t, HStruct):
                    intf = StructIntf(t, self._instantiateFieldFn)
                else:
                    intf = self._instantiateFieldFn(self, field)

                self._fieldsToInterfaces[field] = intf
                setattr(self, field.name, intf)

                if isinstance(intf, StructIntf):
                    intf._fieldsToInterfaces = self._fieldsToInterfaces

    def _getSimAgent(self):
        from hwt.interfaces.agents.structIntf import StructIntfAgent
        return StructIntfAgent
